extern crate ctrlc;

use clichess::Player;
use clichess::{PlayingStatus, RecvPositionError, UserRole, EXIT_MSG};
use serde_json::json;
use shakmaty::fen::Fen;
use shakmaty::{Chess, Color, Outcome, Position, Setup};
use std::io;
use std::iter::Iterator;
use std::os::unix::net::UnixStream;
use std::sync::atomic::{AtomicBool, Ordering};
use std::sync::mpsc::{channel, Receiver, TryRecvError};
use std::sync::{Arc, Mutex};
use std::thread;
use std::time;

struct Client {
    player: Player,
    side: Color,
    running: Arc<AtomicBool>,
    input_buffer: Arc<Mutex<String>>,
    server_message_recv: Receiver<String>,
    playing: bool,
}

enum LobbyLoopOutcome {
    ClientLeft,
}

enum PlayingLoopOutcome {
    PlayerLeft,
    Outcome(Outcome),
}

fn main() {
    let running = setupctrlc();
    let username = std::env::args().nth(1).expect("no name given");
    let public_key = std::env::args().nth(2).expect("no public key given");
    //send username and public key to server
    let mut stream = match UnixStream::connect("/tmp/clichess.socket") {
        Ok(sock) => sock,
        Err(_) => {
            println!("clichess daemon is not running.");
            return;
        }
    };
    let username_pubkey_json = json!({
        "username": username,
        "pubkey": public_key,
    });
    clichess::write_to_stream(&mut stream, username_pubkey_json.to_string()).unwrap();
    let input_buffer = setup_input_buffer();
    //start a thread to listen to server messages
    let server_message_recv = setup_server_message_recv(&stream).unwrap();

    let mut client = Client {
        player: Player {
            id: 0,
            username: username,
            public_key: public_key,
            playing_status: PlayingStatus::WaitingInLobby,
            role: UserRole::Spectator,
        },
        side: Color::White,
        running: running.clone(),
        input_buffer: input_buffer.clone(),
        server_message_recv,
        playing: false,
    };

    //get id from server
    let player_id = fetch_message_from_server(&client).parse::<usize>().unwrap();
    println!("got id {} from server", player_id);
    client.player.id = player_id;

    loop {
        lobby_loop(&client, &mut stream);
        match playing_loop(&client, &mut stream) {
            PlayingLoopOutcome::PlayerLeft => {
                println!("Bye");
                break;
            }
            PlayingLoopOutcome::Outcome(Outcome::Draw) => println!("Draw game."),
            PlayingLoopOutcome::Outcome(Outcome::Decisive { winner }) => {
                if winner == Color::White {
                    println!("White has won the game.")
                } else {
                    println!("Black has won the game.")
                }
            }
        };
    }
}

fn lobby_loop(client: &Client, stream: &mut UnixStream) {
    println!("Sending first refresh to server...");
    clichess::write_to_stream(stream, String::from("refresh")).unwrap();
    println!("Getting list of players from server...");
    let players = fetch_players_from_server(client);
    print_players(client, &players);
    while client.running.load(Ordering::SeqCst) {
        let mut user_input = client.input_buffer.lock().unwrap();
        if user_input.is_empty() {
            continue;
        }
        let mut sent_to_server = String::from(user_input.trim());
        if sent_to_server == "refresh" || sent_to_server == "r" {
            sent_to_server = String::from("refresh");
        }
        println!("Sending {} to server...", sent_to_server);
        clichess::write_to_stream(stream, sent_to_server.clone()).unwrap();
        user_input.clear();
        if sent_to_server == "refresh" {
            let players = fetch_players_from_server(client);
            print_players(client, &players);
        } else {
            let response = fetch_message_from_server(client);
            println!("response from server: {}", response);
        }
    }
}

fn fetch_players_from_server(client: &Client) -> Vec<Player> {
    let data = fetch_message_from_server(client);
    let v: Vec<Player> = serde_json::from_str(&data).unwrap();
    v
}

fn print_players(client: &Client, players: &Vec<Player>) {
    let other_players: Vec<&Player> = players.iter().filter(|p| p.id != client.player.id).collect();
    if other_players.is_empty() {
        println!("No one is connected.");
    }
    for player in other_players {
        let mut message_to_add: String;
        match player.playing_status {
            PlayingStatus::WaitingInLobby => message_to_add = String::default(),
            PlayingStatus::Challenging(id) => {
                message_to_add = String::from(if id == client.player.id {
                    " is challenging you"
                } else {
                    ""
                })
            }
            PlayingStatus::PlayingAgainst(id) => {
                message_to_add = " is playing against ".to_owned();
                message_to_add.push_str(
                    &players
                        .iter()
                        .filter(|p| p.id == id)
                        .last()
                        .expect("id is from the list")
                        .username,
                );
            }
        };
        println!("{}{}", player.username, message_to_add);
    }
}

fn playing_loop(client: &Client, stream: &mut UnixStream) -> PlayingLoopOutcome {
    let mut current_position = fetch_initial_chess_position(&client);
    loop {
        println!(
            "{}",
            clichess::board_representation(&current_position, client.side)
        );
        //check if game is over.
        if current_position.is_game_over() {
            break;
        }
        if clichess::is_player_turn(&client.player, current_position.turn()) {
            //it's the user turn, taking user input
            let input = read_user_input(client);
            clichess::write_to_stream(stream, String::from(input.trim())).unwrap();
            if input.trim() == EXIT_MSG {
                break;
            }
        }
        //update position after playing.
        match get_current_position(client) {
            Ok(position) => current_position = position,
            Err(_) => {
                clichess::write_to_stream(stream, String::from(EXIT_MSG)).unwrap();
                break;
            }
        };
    }
    match current_position.outcome() {
        None => PlayingLoopOutcome::PlayerLeft,
        Some(outcome) => PlayingLoopOutcome::Outcome(outcome),
    }
}

fn setup_input_buffer() -> Arc<Mutex<String>> {
    let buf = Arc::new(Mutex::new(String::new()));
    let buf2 = buf.clone();

    thread::spawn(move || {
        loop {
            let mut buffer = String::new();
            //wait for user input
            io::stdin().read_line(&mut buffer).unwrap();
            {
                let mut user_input = buf2.lock().unwrap();
                if user_input.is_empty() {
                    *user_input = buffer;
                }
            }
        }
    });
    buf
}

fn setupctrlc() -> Arc<AtomicBool> {
    let running = Arc::new(AtomicBool::new(true));
    let r = running.clone();
    ctrlc::set_handler(move || {
        r.store(false, Ordering::SeqCst);
    })
    .expect("Error setting Ctrl-C handler");
    running
}

fn setup_server_message_recv(stream: &UnixStream) -> io::Result<Receiver<String>> {
    let (sender, receiver) = channel();
    let thread_stream = stream.try_clone()?;

    thread::spawn(move || {
        loop {
            //wait for server message
            let buffer =
                clichess::read_line_from_stream(&thread_stream).expect("Error message from server");
            sender.send(buffer).unwrap();
        }
    });
    Ok(receiver)
}

fn read_user_input(client: &Client) -> String {
    //clear input before waiting for a new one
    {
        let mut user_input = client.input_buffer.lock().unwrap();
        user_input.clear();
    }
    let mut input = String::new();
    while client.running.load(Ordering::SeqCst) {
        thread::sleep(time::Duration::from_millis(10));
        {
            let user_input = client.input_buffer.lock().unwrap();
            if user_input.is_empty() {
                continue;
            } else {
                input = user_input.clone();
                break;
            }
        }
    }
    if client.running.load(Ordering::SeqCst) {
        input
    } else {
        String::from(EXIT_MSG)
    }
}

//wait for next position from server, then return the current board.
fn get_current_position(client: &Client) -> Result<Chess, RecvPositionError> {
    let response = fetch_message_from_server(client);
    if response.is_empty() {
        Err(RecvPositionError::UserCanceledError)
    } else {
        Ok(parse_position(&response))
    }
}

fn fetch_message_from_server(client: &Client) -> String {
    let mut response = String::new();
    while client.running.load(Ordering::SeqCst) {
        thread::sleep(time::Duration::from_millis(10));
        {
            let server_response_recv = &client.server_message_recv;
            match server_response_recv.try_recv() {
                Ok(msg) => {
                    response = msg;
                    break;
                }
                Err(TryRecvError::Disconnected) => println!("Error: server disconnected."),
                Err(TryRecvError::Empty) => continue,
            }
        }
    }
    response
}

fn parse_position(string: &str) -> Chess {
    let setup: Fen = string.trim().parse().expect("Invalid message from server.");
    let position: Chess = setup.position().expect("Invalid message from server.");
    position
}

fn fetch_initial_chess_position(client: &Client) -> Chess {
    parse_position(&fetch_message_from_server(client))
}
